--- src/VBox/Frontends/VBoxHeadless/VBoxHeadless.cpp.orig	2010-03-26 04:56:06.000000000 +0900
+++ src/VBox/Frontends/VBoxHeadless/VBoxHeadless.cpp	2010-04-13 20:56:33.000000000 +0900
@@ -47,6 +47,17 @@
 #include <VBox/err.h>
 #include <VBox/VBoxVideo.h>
 
+#ifdef VBOX_WITH_LIBVNC
+#ifdef VBOX_FFMPEG
+#undef VBOX_FFMPEG
+#endif
+#include "VNCFramebuffer.h"
+#ifdef __FreeBSD__
+#include <sys/types.h>
+#include <unistd.h>
+#endif /* __FreeBSD__ */
+#endif /* VBOX_WITH_LIBVNC */
+
 #ifdef VBOX_FFMPEG
 #include <cstdlib>
 #include <cerrno>
@@ -356,6 +367,13 @@
 {
     RTPrintf("Usage:\n"
              "   -s, -startvm, --startvm <name|uuid>   Start given VM (required argument)\n"
+#ifdef VBOX_WITH_LIBVNC
+             "   -v, -vnc, --vnc on|off                Enable (default) or disable the VNC Server\n"
+             "   -a, -vncaddress, --vncaddress <ip>    IP address the VNC server will bind to\n"
+             "   -p, -vncport, --vncport <port>        Port number the VNC server will bind to\n"
+             "   -k, -vnckeymap, --vnckeymap <keymap>  Keyboard mapfile (default: builtin US)\n"
+             "   -S, -vncsecret, --vncsecret <secret>  VNC Authentication secret\n"
+#endif /* VBOX_WITH_LIBVNC */
 #ifdef VBOX_WITH_VRDP
              "   -v, -vrdp, --vrdp on|off|config       Enable (default) or disable the VRDP\n"
              "                                         server or don't change the setting\n"
@@ -427,6 +445,13 @@
  */
 extern "C" DECLEXPORT (int) TrustedMain (int argc, char **argv, char **envp)
 {
+#ifdef VBOX_WITH_LIBVNC
+    unsigned int vncPort = 5900;
+    const char *vncAddress = NULL;
+    const char *vncEnabled = NULL;
+    const char *vncKeymap = NULL;
+    const char *vncSecret = NULL;
+#endif /* VBOX_WITH_LIBVNC */
 #ifdef VBOX_WITH_VRDP
     const char *vrdpPort = NULL;
     const char *vrdpAddress = NULL;
@@ -481,6 +506,18 @@
     {
         { "-startvm", 's', RTGETOPT_REQ_STRING },
         { "--startvm", 's', RTGETOPT_REQ_STRING },
+#ifdef VBOX_WITH_LIBVNC
+        { "-vncport", 'p', RTGETOPT_REQ_UINT32 },
+        { "--vncport", 'p', RTGETOPT_REQ_UINT32 },
+        { "-vncaddress", 'a', RTGETOPT_REQ_STRING },
+        { "--vncaddress", 'a', RTGETOPT_REQ_STRING },
+        { "-vnc", 'v', RTGETOPT_REQ_STRING },
+        { "--vnc", 'v', RTGETOPT_REQ_STRING },
+        { "-vnckeymap", 'k', RTGETOPT_REQ_STRING },
+        { "--vnckeymap", 'k', RTGETOPT_REQ_STRING },
+        { "-vncsecret", 'S', RTGETOPT_REQ_STRING },
+        { "--vncsecret", 'S', RTGETOPT_REQ_STRING },
+#endif /* VBOX_WITH_LIBVNC */
 #ifdef VBOX_WITH_VRDP
         { "-vrdpport", 'p', RTGETOPT_REQ_STRING },
         { "--vrdpport", 'p', RTGETOPT_REQ_STRING },
@@ -532,6 +569,23 @@
                 if (!id)
                     name = ValueUnion.psz;
                 break;
+#ifdef VBOX_WITH_LIBVNC
+            case 'p':
+                vncPort = ValueUnion.u32;
+                break;
+            case 'a':
+                vncAddress = ValueUnion.psz;
+                break;
+            case 'v':
+                vncEnabled = ValueUnion.psz;
+                break;
+            case 'k':
+                vncKeymap = ValueUnion.psz;
+                break;
+            case 'S':
+                vncSecret = ValueUnion.psz;
+                break;
+#endif /* VBOX_WITH_LIBVNC */
 #ifdef VBOX_WITH_VRDP
             case 'p':
                 vrdpPort = ValueUnion.psz;
@@ -609,6 +663,16 @@
                 return -1;
         }
     }
+#ifdef VBOX_WITH_LIBVNC
+#ifdef __FreeBSD__
+    if (vncEnabled != NULL &&
+	(!strcmp(vncEnabled, "off") || !strcmp(vncEnabled, "disable"))) {
+	setproctitle("VM: %s Port: %d Auth: %s", (name != NULL) ? name : "none", 0, "off");
+    } else {
+	setproctitle("VM: %s Port: %d Auth: %s", (name != NULL) ? name : "none", vncPort, (vncSecret != NULL) ? "on" : "off");
+    }
+#endif /* __FreeBSD__ */
+#endif /* VBOX_WITH_LIBVNC */
 
 #ifdef VBOX_FFMPEG
     if (ulFrameWidth < 512 || ulFrameWidth > 2048 || ulFrameWidth % 2)
@@ -855,6 +919,48 @@
                 break;
         }
 
+#ifdef VBOX_WITH_LIBVNC
+	BOOL fVNCEnabled = true;
+	VBoxVNCFB *pVNCFramebuffer = NULL;
+        unsigned uScreenId = 0;
+
+	if (vncEnabled != NULL)
+	{
+            /* -vnc on|off */
+            if (!strcmp(vncEnabled, "off") || !strcmp(vncEnabled, "disable"))
+                fVNCEnabled = false;
+            else if (strcmp(vncEnabled, "on") && strcmp(vncEnabled, "enable"))
+            {
+                RTPrintf("-vnc requires an argument (on|off)\n");
+                break;
+            }
+        }
+
+        if (fVNCEnabled)
+        {
+            Log (("VBoxHeadless: Enabling VNC server...\n"));
+
+            pVNCFramebuffer = new VBoxVNCFB ();
+            if (!pVNCFramebuffer)
+            {
+                RTPrintf("Error: failed to allocate framebuffer\n");
+                break;
+            }
+            if (!pVNCFramebuffer->loadKeymap(vncKeymap)) {
+                RTPrintf("Error: failed to load keymap\n");
+                break;
+	    }
+            pVNCFramebuffer->setAuthInfo(vncSecret);
+            if (!pVNCFramebuffer->init(name, vncPort, vncAddress))
+            {
+                RTPrintf("Error: failed to initialize framebuffer\n");
+                break;
+            }
+            pVNCFramebuffer->setConsole(console);
+            pVNCFramebuffer->AddRef();
+            display->SetFramebuffer(uScreenId, pVNCFramebuffer);
+        }
+#endif /* VBOX_WITH_LIBVNC */
 #ifdef VBOX_WITH_VRDP
         /* default is to enable the RDP server (backward compatibility) */
         BOOL fVRDPEnable = true;
@@ -949,6 +1055,14 @@
 
         Log (("VBoxHeadless: event loop has terminated...\n"));
 
+#ifdef VBOX_WITH_LIBVNC
+        if (pVNCFramebuffer)
+        {
+            pVNCFramebuffer->Release ();
+            Log(("Released framebuffer\n"));
+            pVNCFramebuffer = NULL;
+        }
+#endif /* VBOX_WITH_LIBVNC */
 #ifdef VBOX_FFMPEG
         if (pFramebuffer)
         {
