--- src/VBox/Frontends/VBoxHeadless/VNCFramebuffer.cpp.orig	1970-01-01 09:00:00.000000000 +0900
+++ src/VBox/Frontends/VBoxHeadless/VNCFramebuffer.cpp	2010-04-21 01:02:08.000000000 +0900
@@ -0,0 +1,516 @@
+/*
+ * VNCFramebuffer with LibVNCServer
+ *   designed for FreeNAS 0.7.1
+ * Created by Daisuke Aoyama <aoyama@peach.ne.jp>
+ *
+ * AltGr support code is referred to Hubert Tournier's patch
+ */
+
+#include <VBox/com/com.h>
+#include <VBox/com/string.h>
+#include <VBox/com/Guid.h>
+#include <VBox/com/ErrorInfo.h>
+#include <VBox/com/EventQueue.h>
+#include <VBox/com/VirtualBox.h>
+
+#include <iprt/stream.h>
+#include <iprt/env.h>
+
+#include <ctype.h>
+#include <string.h>
+#include <sys/types.h>
+#include <sys/socket.h>
+#include <netinet/in.h>
+#include <arpa/inet.h>
+
+using namespace com;
+
+#define LOG_GROUP LOG_GROUP_GUI
+#include <VBox/err.h>
+#include <VBox/log.h>
+
+#include "VBoxHeadless.h"
+#include "VNCFramebuffer.h"
+#include "keymap.h"
+
+#if defined(VBOX_WITH_XPCOM)
+NS_IMPL_ISUPPORTS1_CI(VBoxVNCFB, IFramebuffer)
+NS_DECL_CLASSINFO(VBoxVNCFB)
+#endif
+
+static rfbBool dopwcheck(rfbClientPtr cl, const char* response, int len);
+static void doptr(int buttonMask, int x, int y, rfbClientPtr cl);
+static void dokbd(rfbBool down, rfbKeySym keySym, rfbClientPtr cl);
+
+VBoxVNCFB::VBoxVNCFB()
+{
+    int rc;
+    LogFlow(("VBoxFNCFB::VBoxVNCFB\n"));
+    rc = RTCritSectInit(&m_UpdateLock);
+    AssertMsg(rc == VINF_SUCCESS, ("Error from RTCritSectInit!\n"));
+    m_Screen = NULL;
+    m_secret = NULL;
+    m_Keyboard = NULL;
+    m_Mouse = NULL;
+    m_ptrx = m_ptry = -1;
+}
+
+VBoxVNCFB::~VBoxVNCFB()
+{
+    LogFlow(("VBoxVNCFB::~VBoxVNCFB\n"));
+    if (m_Screen != NULL) {
+	rfbShutdownServer(m_Screen, TRUE);
+	free((void*)m_Screen->desktopName);
+	free((void*)m_Screen->frameBuffer);
+	rfbScreenCleanup(m_Screen);
+	m_Screen = NULL;
+    }
+    RTCritSectDelete(&m_UpdateLock);
+}
+
+bool VBoxVNCFB::init(const char *name, unsigned int port, const char *address)
+{
+    int width = 640, height = 480, bps = 8, spp = 3, bpp =4;
+    int rmax = 255, gmax = 255, bmax = 255;
+    int rshift = 16, gshift = 8, bshift = 0;
+    int dname_len;
+    char *dname;
+
+    m_Screen = rfbGetScreen(NULL, NULL, width, height, bps, spp, bpp);
+    dname_len = 64 + (name != NULL ? strlen(name) : 0) + 1;
+    dname = (char*)malloc(dname_len);
+    if (dname == NULL)
+	return false;
+    if (name != NULL) {
+	snprintf(dname, dname_len, "VBox (%s)", name);
+    } else {
+	snprintf(dname, dname_len, "VBox");
+    }
+    m_Screen->desktopName = dname;
+    m_Screen->port = port;
+    if (address == NULL
+	|| strcmp(address, "*") == 0
+	|| strcmp(address, "0.0.0.0") == 0) {
+	m_Screen->listenInterface = htonl(INADDR_ANY);
+    } else {
+	m_Screen->listenInterface = inet_addr(address);
+    }
+    m_Screen->frameBuffer = (char*)malloc(width * height * bpp);
+    if (m_Screen->frameBuffer == NULL)
+	return false;
+    memset(m_Screen->frameBuffer, 0, (width * height * bpp));
+    m_Screen->ptrAddEvent = doptr;
+    m_Screen->kbdAddEvent = dokbd;
+    m_Screen->screenData = (void*)this;
+    m_Screen->authPasswdData = (void*)m_secret;
+    m_Screen->passwordCheck = dopwcheck;
+    rfbInitServer(m_Screen);
+    m_Screen->serverFormat.redMax = rmax;
+    m_Screen->serverFormat.greenMax = gmax;
+    m_Screen->serverFormat.blueMax = bmax;
+    m_Screen->serverFormat.redShift = rshift;
+    m_Screen->serverFormat.greenShift = gshift;
+    m_Screen->serverFormat.blueShift = bshift;
+    rfbRunEventLoop(m_Screen, -1, TRUE);
+
+    return true;
+}
+
+bool VBoxVNCFB::loadKeymap(const char *keymap)
+{
+    if (load_keymap_table(keymap) != 0) {
+	return false;
+    }
+    return true;
+}
+
+void VBoxVNCFB::setAuthInfo(const char *secret)
+{
+    m_secret = secret;
+}
+
+void VBoxVNCFB::setConsole(ComPtr<IConsole> &Console)
+{
+    Console->COMGETTER(Keyboard)(m_Keyboard.asOutParam());
+    AssertMsg(m_Keyboard, ("Error keyboard\n"));
+    Console->COMGETTER(Mouse)(m_Mouse.asOutParam());
+    AssertMsg(m_Mouse, ("Error mouse\n"));
+}
+
+static rfbBool dopwcheck(rfbClientPtr cl, const char* response, int len)
+{
+    VBoxVNCFB *fb = (VBoxVNCFB*)cl->screen->screenData;
+    char *secret = (char*)fb->getAuthInfo();
+
+#if 0
+    printf("secret=%s\n", secret);
+#endif
+    if (secret == NULL || secret[0] == '\0')
+	return FALSE;
+    rfbEncryptBytes(cl->authChallenge, secret);
+    if (memcmp(cl->authChallenge, response, len) != 0) {
+	rfbErr("authentication failed from %s\n", cl->host);
+	return FALSE;
+    }
+    return TRUE;
+}
+
+static void doptr(int buttonMask, int x, int y, rfbClientPtr cl)
+{
+    VBoxVNCFB *fb = (VBoxVNCFB*)cl->screen->screenData;
+    ComPtr<IMouse> Mouse = fb->getMouse();
+    BOOL absoluteSupported;
+    int dx = 0, dy = 0, dz = 0, buttons = 0;
+
+#if 0
+    printf("ptr: x=%d, y=%d, mask=%d\n", x, y, buttonMask);
+#endif
+    if (buttonMask & 0x01)
+	buttons |= MouseButtonState_LeftButton;
+    if (buttonMask & 0x02)
+	buttons |= MouseButtonState_MiddleButton;
+    if (buttonMask & 0x04)
+	buttons |= MouseButtonState_RightButton;
+    if (buttonMask & 0x08)
+	dz = -1;
+    if (buttonMask & 0x10)
+	dz = 1;
+
+    Mouse->GetAbsoluteSupported(&absoluteSupported);
+    if (absoluteSupported) {
+	Mouse->PutMouseEventAbsolute(x, y, dz, 0, buttons);
+    } else {
+	int ptrx, ptry;
+	fb->getMousePtr(ptrx, ptry);
+	if (ptrx != -1 && ptry != -1) {
+	    dx = ((x - ptrx) * 160) / 100;
+	    dy = ((y - ptry) * 160) / 100;
+	    Mouse->PutMouseEvent(dx, dy, dz, 0, buttons);
+	}
+	fb->setMousePtr(x, y);
+    }
+}
+
+static void dokbd(rfbBool down, rfbKeySym keySym, rfbClientPtr cl)
+{
+    VBoxVNCFB *fb = (VBoxVNCFB*)cl->screen->screenData;
+    ComPtr<IKeyboard> Keyboard = fb->getKeyboard();
+    int scancode, modifier, scancode2;
+
+#if 0
+    printf("down=%d, Sym=%x\n", down ? 1 : 0, keySym);
+#endif
+    if (keySym < 0x100 && isalpha(keySym)) {
+	keySym = tolower(keySym);
+    }
+    switch (keySym) {
+    case XK_Alt_R:
+	if (!need_alternate_graphic()) {
+	    scancode = keysym2scancode(XK_Alt_L);
+	    scancode &= 0x7f;
+	    Keyboard->PutScancode(0xe0);
+	    Keyboard->PutScancode(!down ? scancode | 0x80 : scancode);
+	}
+	break;
+    case XK_Control_R:
+	scancode = keysym2scancode(XK_Control_L);
+	scancode &= 0x7f;
+	Keyboard->PutScancode(0xe0);
+	Keyboard->PutScancode(!down ? scancode | 0x80 : scancode);
+	break;
+    case XK_Print:
+	scancode = 0x2a;
+	Keyboard->PutScancode(0xe0);
+	Keyboard->PutScancode(!down ? scancode | 0x80 : scancode);
+	scancode = 0x37;
+	Keyboard->PutScancode(0xe0);
+	Keyboard->PutScancode(!down ? scancode | 0x80 : scancode);
+	break;
+    case XK_Break:
+	scancode = keysym2scancode(XK_Scroll_Lock);
+	scancode &= 0x7f;
+	Keyboard->PutScancode(0xe0);
+	Keyboard->PutScancode(!down ? scancode | 0x80 : scancode);
+	break;
+#if 1
+    case XK_KP_0:
+	keySym = '0'; goto normal_scancode;
+    case XK_KP_1:
+	keySym = '1'; goto normal_scancode;
+    case XK_KP_2:
+	keySym = '2'; goto normal_scancode;
+    case XK_KP_3:
+	keySym = '3'; goto normal_scancode;
+    case XK_KP_4:
+	keySym = '4'; goto normal_scancode;
+    case XK_KP_5:
+	keySym = '5'; goto normal_scancode;
+    case XK_KP_6:
+	keySym = '6'; goto normal_scancode;
+    case XK_KP_7:
+	keySym = '7'; goto normal_scancode;
+    case XK_KP_8:
+	keySym = '8'; goto normal_scancode;
+    case XK_KP_9:
+	keySym = '9'; goto normal_scancode;
+    case XK_KP_Decimal:
+	keySym = '.'; goto normal_scancode;
+#endif
+    default:
+	;
+    normal_scancode:
+	scancode = keysym2scancode_modifier(keySym, &modifier);
+	if (scancode < 0) {
+	    //printf("unknown code %x, %d\n", keySym, down);
+	} else {
+#if 0
+	    printf("scan code %x => %x, %d, %d\n", keySym, scancode, down,
+		   modifier);
+#endif
+	    if ((modifier & S2KMOD_ALT)
+		&& need_alternate_graphic()) {
+		scancode2 = keysym2scancode(XK_Alt_L);
+		scancode2 &= 0x7f;
+		Keyboard->PutScancode(0xe0);
+		Keyboard->PutScancode(!down ? scancode2 | 0x80 : scancode2);
+	    }
+	    scancode &= 0x7f;
+	    Keyboard->PutScancode(!down ? scancode | 0x80 : scancode);
+	}
+    }
+}
+
+STDMETHODIMP VBoxVNCFB::COMGETTER(Width)(ULONG *width)
+{
+    LogFlow(("VBoxVNCFB::GetWidth\n"));
+    if (!width)
+        return E_INVALIDARG;
+    *width = (ULONG)m_Screen->width;
+    return S_OK;
+}
+
+STDMETHODIMP VBoxVNCFB::COMGETTER(Height)(ULONG *height)
+{
+    LogFlow(("VBoxVNCFB::GetHeight\n"));
+    if (!height)
+        return E_INVALIDARG;
+    *height = (ULONG)m_Screen->height;
+    return S_OK;
+}
+
+STDMETHODIMP VBoxVNCFB::Lock()
+{
+    LogFlow(("VBoxVNCFB::Lock\n"));
+    RTCritSectEnter(&m_UpdateLock);
+    return S_OK;
+}
+
+STDMETHODIMP VBoxVNCFB::Unlock()
+{
+    LogFlow(("VBoxVNCFB::Unlock\n"));
+    RTCritSectLeave(&m_UpdateLock);
+    return S_OK;
+}
+
+STDMETHODIMP VBoxVNCFB::COMGETTER(Address)(BYTE **address)
+{
+    LogFlow(("VBoxVNCFB::GetAddress\n"));
+    if (!address)
+        return E_INVALIDARG;
+    *address = (BYTE*)m_Screen->frameBuffer;
+    return S_OK;
+}
+
+STDMETHODIMP VBoxVNCFB::COMGETTER(BitsPerPixel)(ULONG *bitsPerPixel)
+{
+    LogFlow(("VBoxVNCFB::GetBitsPerPixel\n"));
+    if (!bitsPerPixel)
+        return E_INVALIDARG;
+    *bitsPerPixel = (ULONG)m_Screen->bitsPerPixel;
+    return S_OK;
+}
+
+STDMETHODIMP VBoxVNCFB::COMGETTER(BytesPerLine)(ULONG *bytesPerLine)
+{
+    LogFlow(("VBoxVNCFB::GetBytesPerLine\n"));
+    if (!bytesPerLine)
+        return E_INVALIDARG;
+    *bytesPerLine = (ULONG)(m_Screen->width * (m_Screen->bitsPerPixel / 8));
+    return S_OK;
+}
+
+STDMETHODIMP VBoxVNCFB::COMGETTER(PixelFormat)(ULONG *pixelFormat)
+{
+    if (!pixelFormat)
+        return E_POINTER;
+    *pixelFormat = FramebufferPixelFormat_FOURCC_RGB;
+    return S_OK;
+}
+
+STDMETHODIMP VBoxVNCFB::COMGETTER(UsesGuestVRAM)(BOOL *usesGuestVRAM)
+{
+    if (!usesGuestVRAM)
+        return E_POINTER;
+    *usesGuestVRAM = FALSE;
+    return S_OK;
+}
+
+STDMETHODIMP VBoxVNCFB::COMGETTER(HeightReduction)(ULONG *heightReduction)
+{
+    if (!heightReduction)
+        return E_POINTER;
+    *heightReduction = 0;
+    return S_OK;
+}
+
+STDMETHODIMP VBoxVNCFB::COMGETTER(Overlay)(IFramebufferOverlay **aOverlay)
+{
+    if (!aOverlay)
+        return E_POINTER;
+    *aOverlay = 0;
+    return S_OK;
+}
+
+STDMETHODIMP VBoxVNCFB::COMGETTER(WinId)(uint64_t *winId)
+{
+    if (!winId)
+        return E_POINTER;
+    *winId = 0;
+    return S_OK;
+}
+
+STDMETHODIMP VBoxVNCFB::NotifyUpdate(ULONG x, ULONG y, ULONG w, ULONG h)
+{
+    ULONG x2, y2;
+
+    if (w == 0)
+	w = 1;
+    if (h == 0)
+	h = 1;
+    x2 = x + w;
+    y2 = y + h;
+
+    if (x >= (ULONG)m_Screen->width)
+	x = (ULONG)m_Screen->width;
+    if (x2 >= (ULONG)m_Screen->width)
+	x2 = (ULONG)m_Screen->width;
+    if (y >= (ULONG)m_Screen->height)
+	y = (ULONG)m_Screen->height;
+    if (y2 >= (ULONG)m_Screen->height)
+	y2 = (ULONG)m_Screen->height;
+#if 0
+    printf("Update X=%d, Y=%d, W=%d, H=%d =>", x, y, w, h);
+    printf("X=%d, Y=%d, X2=%d, Y2=%d\n", x, y, x2, y2);
+#endif
+    rfbMarkRectAsModified(m_Screen, x, y, x2, y2);
+    return S_OK;
+}
+
+STDMETHODIMP VBoxVNCFB::RequestResize(ULONG aScreenId, ULONG pixelFormat, BYTE *vram,
+                                   ULONG bitsPerPixel, ULONG bytesPerLine,
+                                   ULONG w, ULONG h, BOOL *finished)
+{
+    if (!finished)
+	return E_FAIL;
+#if 0
+    AssertReturn(pixelFormat == FramebufferPixelFormat_FOURCC_RGB, E_INVALIDARG);
+    AssertReturn(bitsPerPixel == 32, E_INVALIDARG);
+#endif
+
+    printf("RequestResize: w=%d, h=%d, bpp=%d, bpl=%d, fmt=%x\n",
+	   w, h, bitsPerPixel, bytesPerLine, pixelFormat);
+
+    int width = w, height = h, bps = 8, spp = 3, bpp =4;
+    int rmax = 255, gmax = 255, bmax = 255;
+    int rshift = 16, gshift = 8, bshift = 0;
+    char *oldfb, *newfb;
+
+    switch (bitsPerPixel) {
+    case 32: /* X888 */
+	bps = 8; spp = 3; bpp = 4;
+	rmax = 255; gmax = 255; bmax = 255;
+	rshift = 16; gshift = 8; bshift = 0;
+	break;
+    case 24: /* 888 */
+	bps = 8; spp = 3; bpp = 4;
+	rmax = 255; gmax = 255; bmax = 255;
+	rshift = 16; gshift = 8; bshift = 0;
+	break;
+    case 16: /* 565 */
+#if 0
+	bps = 5; spp = 3; bpp = 2;
+	rmax = 31; gmax = 63; bmax = 31;
+	rshift = 11; gshift = 5; bshift = 0;
+#else
+	/* use 888 instead of 565 */
+	bps = 8; spp = 3; bpp = 4;
+	rmax = 255; gmax = 255; bmax = 255;
+	rshift = 16; gshift = 8; bshift = 0;
+#endif
+	break;
+    case 8: /* 332 */
+	bps = 3; spp = 3; bpp = 1;
+	rmax = 7; gmax = 7; bmax = 3;
+	rshift = 5; gshift = 2; bshift = 0;
+	break;
+    case 0: /* XXX */
+    default:
+	bps = 8; spp = 3; bpp = 4;
+	rmax = 255; gmax = 255; bmax = 255;
+	rshift = 16; gshift = 8; bshift = 0;
+	break;
+    }
+#if 0
+    printf("bps=%d, spp=%d, bpp=%d, r=%d, g=%d, b=%d\n",
+	   bps, spp, bpp, rmax, gmax, bmax);
+#endif
+
+    oldfb = m_Screen->frameBuffer;
+    newfb = (char*)malloc(width * height * bpp);
+    memset(newfb, 0, (width * height * bpp));
+    rfbNewFramebuffer(m_Screen, newfb, width, height, bps, spp, bpp);
+    m_Screen->serverFormat.redMax = rmax;
+    m_Screen->serverFormat.greenMax = gmax;
+    m_Screen->serverFormat.blueMax = bmax;
+    m_Screen->serverFormat.redShift = rshift;
+    m_Screen->serverFormat.greenShift = gshift;
+    m_Screen->serverFormat.blueShift = bshift;
+    free(oldfb);
+#if 1
+    rfbMarkRectAsModified(m_Screen, 0, 0, m_Screen->width, m_Screen->height);
+#endif
+
+    *finished = true;
+    return S_OK;
+}
+
+STDMETHODIMP VBoxVNCFB::VideoModeSupported(ULONG width, ULONG height, ULONG bpp, BOOL *supported)
+{
+    if (!supported)
+        return E_POINTER;
+    *supported = true;
+#if 0
+    if (bpp != 32)
+	*supported = false;
+#endif
+    return S_OK;
+}
+
+STDMETHODIMP VBoxVNCFB::GetVisibleRegion(BYTE *aRectangles, ULONG aCount, ULONG *aCountCopied)
+{
+    if (!aRectangles)
+        return E_POINTER;
+    return S_OK;
+}
+
+STDMETHODIMP VBoxVNCFB::SetVisibleRegion(BYTE *aRectangles, ULONG aCount)
+{
+    if (!aRectangles)
+        return E_POINTER;
+    return S_OK;
+}
+
+STDMETHODIMP VBoxVNCFB::ProcessVHWACommand(BYTE *pCommand)
+{
+    return  E_NOTIMPL;
+}
