--- src/VBox/Frontends/VBoxHeadless/VNCFramebuffer.h.orig	1970-01-01 09:00:00.000000000 +0900
+++ src/VBox/Frontends/VBoxHeadless/VNCFramebuffer.h	2010-04-13 20:56:33.000000000 +0900
@@ -0,0 +1,84 @@
+/* VNCFramebuffer.h */
+
+#ifndef __H_VNCFRAMEBUFFER
+#define __H_VNCFRAMEBUFFER
+
+#include <rfb/rfb.h>
+#include <rfb/keysym.h>
+
+#include <iprt/critsect.h>
+
+class VBoxVNCDummyServer;
+class VBoxVNCDummyDisplay;
+
+class VBoxVNCFB : VBOX_SCRIPTABLE_IMPL(IFramebuffer)
+{
+public:
+    VBoxVNCFB();
+    virtual ~VBoxVNCFB();
+
+#ifdef RT_OS_WINDOWS
+    STDMETHOD_(ULONG, AddRef)()
+    {
+        return ::InterlockedIncrement (&refcnt);
+    }
+    STDMETHOD_(ULONG, Release)()
+    {
+        long cnt = ::InterlockedDecrement (&refcnt);
+        if (cnt == 0)
+            delete this;
+        return cnt;
+    }
+#endif
+    VBOX_SCRIPTABLE_DISPATCH_IMPL(IFramebuffer)
+
+    NS_DECL_ISUPPORTS
+
+    STDMETHOD(COMGETTER(Width))(ULONG *width);
+    STDMETHOD(COMGETTER(Height))(ULONG *height);
+    STDMETHOD(Lock)();
+    STDMETHOD(Unlock)();
+    STDMETHOD(COMGETTER(Address))(BYTE **address);
+    STDMETHOD(COMGETTER(BitsPerPixel))(ULONG *bitsPerPixel);
+    STDMETHOD(COMGETTER(BytesPerLine))(ULONG *bytesPerLine);
+    STDMETHOD(COMGETTER(PixelFormat))(ULONG *pixelFormat);
+    STDMETHOD(COMGETTER(UsesGuestVRAM))(BOOL *usesGuestVRAM);
+    STDMETHOD(COMGETTER(HeightReduction))(ULONG *heightReduction);
+    STDMETHOD(COMGETTER(Overlay))(IFramebufferOverlay **aOverlay);
+    STDMETHOD(COMGETTER(WinId))(uint64_t *winId);
+
+    STDMETHOD(NotifyUpdate)(ULONG x, ULONG y, ULONG w, ULONG h);
+    STDMETHOD(RequestResize)(ULONG aScreenId, ULONG pixelFormat, BYTE *vram,
+                             ULONG bitsPerPixel, ULONG bytesPerLine,
+                             ULONG w, ULONG h, BOOL *finished);
+    STDMETHOD(VideoModeSupported)(ULONG width, ULONG height, ULONG bpp, BOOL *supported);
+
+    STDMETHOD(GetVisibleRegion)(BYTE *aRectangles, ULONG aCount, ULONG *aCountCopied);
+    STDMETHOD(SetVisibleRegion)(BYTE *aRectangles, ULONG aCount);
+
+    STDMETHOD(ProcessVHWACommand)(BYTE *pCommand);
+
+    bool init(const char *name, unsigned int port, const char *address);
+    bool loadKeymap(const char *keymap);
+    void setAuthInfo(const char *secret);
+    const char *getAuthInfo() { return m_secret; }
+    void setMousePtr(int x, int y) { m_ptrx = x; m_ptry = y; }
+    void getMousePtr(int &x, int &y) { x = m_ptrx; y = m_ptry; }
+    void setConsole(ComPtr<IConsole> &Console);
+    ComPtr<IKeyboard> getKeyboard() { return m_Keyboard; }
+    ComPtr<IMouse> getMouse() { return m_Mouse; }
+
+private:
+    rfbScreenInfoPtr m_Screen;
+    RTCRITSECT m_UpdateLock;
+    const char *m_secret;
+    ComPtr<IKeyboard> m_Keyboard;
+    ComPtr<IMouse> m_Mouse;
+    int m_ptrx, m_ptry;
+
+#ifdef RT_OS_WINDOWS
+    long refcnt;
+#endif
+};
+
+#endif // __H_VNCFRAMEBUFFER
