--- src/VBox/Frontends/VBoxHeadless/keymap.c.orig	1970-01-01 09:00:00.000000000 +0900
+++ src/VBox/Frontends/VBoxHeadless/keymap.c	2010-04-21 01:02:54.000000000 +0900
@@ -0,0 +1,972 @@
+/* convert keysym to scancode */
+/* Created by Daisuke Aoyama <aoyama@peach.ne.jp> */
+/* AltGr support code is referred to Hubert Tournier's patch */
+
+#include <ctype.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+
+#include <rfb/keysym.h>
+#include "keymap.h"
+
+#ifdef VBOX_WITH_LIBVNC
+#define KEYMAPDIR "/usr/local/share/virtualbox-ose/keymaps"
+#else
+#define KEYMAPDIR "/usr/share/syscons/keymaps"
+#endif /* VBOX_WITH_LIBVNC */
+#if 0
+#ifndef DEFAULT_KEYMAP
+#define DEFAULT_KEYMAP "builtin_us"
+#endif /* DEFAULT_KEYMAP */
+#endif
+
+#ifdef TEST
+#define KEYMAPFILE KEYMAPDIR "/us.pc-ctrl.kbd"
+/*
+#define KEYMAPFILE KEYMAPDIR "/jp.106.kbd"
+*/
+#endif /* TEST */
+
+typedef struct
+{
+    char *kbdstr;
+    int keysym;
+} kbdstr2keysym_t;
+
+typedef struct
+{
+    int scancode;
+    int keysym;
+    int modifier;
+} scancode2keysym_t;
+
+#define MAX_TMPBUF 4096
+#define XXX_NOTIMPL(label) 0
+
+static kbdstr2keysym_t kbdstr2keysym[] =
+{
+    { "nop",    0 },
+    { "bs",     XK_BackSpace },
+    { "ht",     XK_Tab },
+    { "",       XK_Linefeed },
+    { "",       XK_Clear },
+    { "cr",     XK_Return },
+    { "",       XK_Pause },
+    { "slock",  XK_Scroll_Lock },
+    { "",       XK_Sys_Req },
+    { "esc",    XK_Escape },
+    { "del",    XK_Delete },
+
+    { "fkey49", XK_Home },
+    { "fkey53", XK_Left },
+    { "fkey50", XK_Up },
+    { "fkey55", XK_Right },
+    { "fkey58", XK_Down },
+    { "",       XK_Prior },
+    { "fkey51", XK_Page_Up },
+    { "",       XK_Next },
+    { "fkey59", XK_Page_Down },
+    { "fkey57", XK_End },
+    { "",       XK_Begin },
+
+    { "",       XK_Select },
+    { "",       XK_Print },
+    { "",       XK_Execute },
+    { "fkey60", XK_Insert },
+    { "",       XK_Undo },
+    { "",       XK_Redo },
+    { "",       XK_Menu },
+    { "",       XK_Find },
+    { "",       XK_Cancel },
+    { "",       XK_Help },
+    { "",       XK_Break },
+    { "",       XK_Mode_switch },
+    { "",       XK_script_switch },
+    { "nlock",  XK_Num_Lock },
+
+    { "lshift", XK_Shift_L },
+    { "rshift", XK_Shift_R },
+    { "lctrl",  XK_Control_L },
+    { "rctrl",  XK_Control_R },
+    { "",       XK_Caps_Lock },
+    { "",       XK_Shift_Lock },
+
+    { "",       XK_Meta_L },
+    { "",       XK_Meta_R },
+    { "lalt",   XK_Alt_L },
+    { "ralt",   XK_Alt_R },
+    { "",       XK_Super_L },
+    { "",       XK_Super_R },
+    { "",       XK_Hyper_L },
+    { "",       XK_Hyper_R },
+
+    { "fkey01", XK_F1 },
+    { "fkey02", XK_F2 },
+    { "fkey03", XK_F3 },
+    { "fkey04", XK_F4 },
+    { "fkey05", XK_F5 },
+    { "fkey06", XK_F6 },
+    { "fkey07", XK_F7 },
+    { "fkey08", XK_F8 },
+    { "fkey09", XK_F9 },
+    { "fkey10", XK_F10 },
+    { "fkey11", XK_F11 },
+    { "",       XK_L1 },
+    { "fkey12", XK_F12 },
+    { "",       XK_L2 },
+    { "fkey13", XK_F13 },
+    { "",       XK_L3 },
+    { "fkey14", XK_F14 },
+    { "",       XK_L4 },
+    { "fkey15", XK_F15 },
+    { "",       XK_L5 },
+    { "fkey16", XK_F16 },
+    { "",       XK_L6 },
+    { "fkey17", XK_F17 },
+    { "",       XK_L7 },
+    { "fkey18", XK_F18 },
+    { "",       XK_L8 },
+    { "fkey19", XK_F19 },
+    { "",       XK_L9 },
+    { "fkey20", XK_F20 },
+    { "",       XK_L10 },
+    { "fkey21", XK_F21 },
+    { "",       XK_R1 },
+    { "fkey22", XK_F22 },
+    { "",       XK_R2 },
+    { "fkey23", XK_F23 },
+    { "",       XK_R3 },
+    { "fkey24", XK_F24 },
+    { "",       XK_R4 },
+    { "",       XK_F25 },
+    { "",       XK_R5 },
+    { "",       XK_F26 },
+    { "",       XK_R6 },
+    { "",       XK_F27 },
+    { "",       XK_R7 },
+    { "",       XK_F28 },
+    { "",       XK_R8 },
+    { "",       XK_F29 },
+    { "",       XK_R9 },
+    { "",       XK_F30 },
+    { "",       XK_R10 },
+    { "",       XK_F31 },
+    { "",       XK_R11 },
+    { "",       XK_F32 },
+    { "",       XK_R12 },
+    { "",       XK_F33 },
+    { "",       XK_R13 },
+    { "",       XK_F34 },
+    { "",       XK_R14 },
+    { "",       XK_F35 },
+    { "",       XK_R15 },
+
+#if 0
+    { "NUM_0",   XK_KP_0 },
+    { "NUM_1",   XK_KP_1 },
+    { "NUM_2",   XK_KP_2 },
+    { "NUM_3",   XK_KP_3 },
+    { "NUM_4",   XK_KP_4 },
+    { "NUM_5",   XK_KP_5 },
+    { "NUM_6",   XK_KP_6 },
+    { "NUM_7",   XK_KP_7 },
+    { "NUM_8",   XK_KP_8 },
+    { "NUM_9",   XK_KP_9 },
+
+    { "NUM_+",   XK_KP_Add },
+    { "NUM_-",   XK_KP_Subtract },
+    { "NUM_*",   XK_KP_Multiply },
+    { "NUM_/",   XK_KP_Divide },
+    { "NUM_.",   XK_KP_Decimal },
+    { "NUM_del", XK_KP_Delete },
+    { "NUM_cr",  XK_KP_Enter },
+
+    { "NUM_fkey49", XK_KP_Home },
+    { "NUM_fkey53", XK_KP_Left },
+    { "NUM_fkey50", XK_KP_Up },
+    { "NUM_fkey55", XK_KP_Right },
+    { "NUM_fkey58", XK_KP_Down },
+    { "",           XK_KP_Prior },
+    { "NUM_fkey51", XK_KP_Page_Up },
+    { "",           XK_KP_Next },
+    { "NUM_fkey59", XK_KP_Page_Down },
+    { "NUM_fkey57", XK_KP_End },
+    { "",           XK_KP_Begin },
+    { "NUM_fkey60", XK_KP_Insert },
+#endif
+
+    { NULL, -1 },
+};
+
+/* generate from /usr/share/syscons/keymaps/us.pc-ctrl.kbd */
+static scancode2keysym_t us_scancode2keysym [] =
+{
+    {   0, 0 },
+    {   0, 0 },
+    {   0, 0 },
+    {   1, XK_Escape },
+    {   1, XK_Escape },
+    {   1, XK_Escape },
+    {   2, '1' },
+    {   2, '!' },
+    {   2, '1' },
+    {   3, '2' },
+    {   3, '@' },
+    {   3, '2' },
+    {   4, '3' },
+    {   4, '#' },
+    {   4, '3' },
+    {   5, '4' },
+    {   5, '$' },
+    {   5, '4' },
+    {   6, '5' },
+    {   6, '%' },
+    {   6, '5' },
+    {   7, '6' },
+    {   7, '^' },
+    {   7, '6' },
+    {   8, '7' },
+    {   8, '&' },
+    {   8, '7' },
+    {   9, '8' },
+    {   9, '*' },
+    {   9, '8' },
+    {  10, '9' },
+    {  10, '(' },
+    {  10, '9' },
+    {  11, '0' },
+    {  11, ')' },
+    {  11, '0' },
+    {  12, '-' },
+    {  12, '_' },
+    {  12, '-' },
+    {  13, '=' },
+    {  13, '+' },
+    {  13, '=' },
+    {  14, XK_BackSpace },
+    {  14, XK_BackSpace },
+    {  14, XK_BackSpace },
+    {  15, XK_Tab },
+    {  15, XXX_NOTIMPL(btab) },
+    {  15, XK_Tab },
+    {  16, 'q' },
+    {  16, 'Q' },
+    {  16, 'q' },
+    {  17, 'w' },
+    {  17, 'W' },
+    {  17, 'w' },
+    {  18, 'e' },
+    {  18, 'E' },
+    {  18, 'e' },
+    {  19, 'r' },
+    {  19, 'R' },
+    {  19, 'r' },
+    {  20, 't' },
+    {  20, 'T' },
+    {  20, 't' },
+    {  21, 'y' },
+    {  21, 'Y' },
+    {  21, 'y' },
+    {  22, 'u' },
+    {  22, 'U' },
+    {  22, 'u' },
+    {  23, 'i' },
+    {  23, 'I' },
+    {  23, 'i' },
+    {  24, 'o' },
+    {  24, 'O' },
+    {  24, 'o' },
+    {  25, 'p' },
+    {  25, 'P' },
+    {  25, 'p' },
+    {  26, '[' },
+    {  26, '{' },
+    {  26, '[' },
+    {  27, ']' },
+    {  27, '}' },
+    {  27, ']' },
+    {  28, XK_Return },
+    {  28, XK_Return },
+    {  28, XK_Return },
+    {  29, XK_Control_L },
+    {  29, XK_Control_L },
+    {  29, XK_Control_L },
+    {  30, 'a' },
+    {  30, 'A' },
+    {  30, 'a' },
+    {  31, 's' },
+    {  31, 'S' },
+    {  31, 's' },
+    {  32, 'd' },
+    {  32, 'D' },
+    {  32, 'd' },
+    {  33, 'f' },
+    {  33, 'F' },
+    {  33, 'f' },
+    {  34, 'g' },
+    {  34, 'G' },
+    {  34, 'g' },
+    {  35, 'h' },
+    {  35, 'H' },
+    {  35, 'h' },
+    {  36, 'j' },
+    {  36, 'J' },
+    {  36, 'j' },
+    {  37, 'k' },
+    {  37, 'K' },
+    {  37, 'k' },
+    {  38, 'l' },
+    {  38, 'L' },
+    {  38, 'l' },
+    {  39, ';' },
+    {  39, ':' },
+    {  39, ';' },
+    {  40, '\'' },
+    {  40, '"' },
+    {  40, '\'' },
+    {  41, '`' },
+    {  41, '~' },
+    {  41, '`' },
+    {  42, XK_Shift_L },
+    {  42, XK_Shift_L },
+    {  42, XK_Shift_L },
+    {  43, '\\' },
+    {  43, '|' },
+    {  43, '\\' },
+    {  44, 'z' },
+    {  44, 'Z' },
+    {  44, 'z' },
+    {  45, 'x' },
+    {  45, 'X' },
+    {  45, 'x' },
+    {  46, 'c' },
+    {  46, 'C' },
+    {  46, 'c' },
+    {  47, 'v' },
+    {  47, 'V' },
+    {  47, 'v' },
+    {  48, 'b' },
+    {  48, 'B' },
+    {  48, 'b' },
+    {  49, 'n' },
+    {  49, 'N' },
+    {  49, 'n' },
+    {  50, 'm' },
+    {  50, 'M' },
+    {  50, 'm' },
+    {  51, ',' },
+    {  51, '<' },
+    {  51, ',' },
+    {  52, '.' },
+    {  52, '>' },
+    {  52, '.' },
+    {  53, '/' },
+    {  53, '?' },
+    {  53, '/' },
+    {  54, XK_Shift_R },
+    {  54, XK_Shift_R },
+    {  54, XK_Shift_R },
+    {  55, XK_KP_Multiply },
+    {  55, XK_KP_Multiply },
+    {  55, XK_KP_Multiply },
+    {  56, XK_Alt_L },
+    {  56, XK_Alt_L },
+    {  56, XK_Alt_L },
+    {  57, ' ' },
+    {  57, ' ' },
+    {  57, XXX_NOTIMPL(nul) },
+    {  58, XK_Control_L },
+    {  58, XK_Control_L },
+    {  58, XK_Control_L },
+    {  59, XK_F1 },
+    {  59, XK_F13 },
+    {  59, XXX_NOTIMPL(scr01) },
+    {  60, XK_F2 },
+    {  60, XK_F14 },
+    {  60, XXX_NOTIMPL(scr02) },
+    {  61, XK_F3 },
+    {  61, XK_F15 },
+    {  61, XXX_NOTIMPL(scr03) },
+    {  62, XK_F4 },
+    {  62, XK_F16 },
+    {  62, XXX_NOTIMPL(scr04) },
+    {  63, XK_F5 },
+    {  63, XK_F17 },
+    {  63, XXX_NOTIMPL(scr05) },
+    {  64, XK_F6 },
+    {  64, XK_F18 },
+    {  64, XXX_NOTIMPL(scr06) },
+    {  65, XK_F7 },
+    {  65, XK_F19 },
+    {  65, XXX_NOTIMPL(scr07) },
+    {  66, XK_F8 },
+    {  66, XK_F20 },
+    {  66, XXX_NOTIMPL(scr08) },
+    {  67, XK_F9 },
+    {  67, XK_F21 },
+    {  67, XXX_NOTIMPL(scr09) },
+    {  68, XK_F10 },
+    {  68, XK_F22 },
+    {  68, XXX_NOTIMPL(scr10) },
+    {  69, XK_Num_Lock },
+    {  69, XK_Num_Lock },
+    {  69, XK_Num_Lock },
+    {  70, XK_Scroll_Lock },
+    {  70, XK_Scroll_Lock },
+    {  70, XK_Scroll_Lock },
+    {  71, XK_Home },
+    {  71, XK_KP_7 },
+    {  71, XK_KP_7 },
+    {  72, XK_Up },
+    {  72, XK_KP_8 },
+    {  72, XK_KP_8 },
+    {  73, XK_Page_Up },
+    {  73, XK_KP_9 },
+    {  73, XK_KP_9 },
+    {  74, XXX_NOTIMPL(fkey52) },
+    {  74, XK_KP_Subtract },
+    {  74, XK_KP_Subtract },
+    {  75, XK_Left },
+    {  75, XK_KP_4 },
+    {  75, XK_KP_4 },
+    {  76, XXX_NOTIMPL(fkey54) },
+    {  76, XK_KP_5 },
+    {  76, XK_KP_5 },
+    {  77, XK_Right },
+    {  77, XK_KP_6 },
+    {  77, XK_KP_6 },
+    {  78, XXX_NOTIMPL(fkey56) },
+    {  78, XK_KP_Add },
+    {  78, XK_KP_Add },
+    {  79, XK_End },
+    {  79, XK_KP_1 },
+    {  79, XK_KP_1 },
+    {  80, XK_Down },
+    {  80, XK_KP_2 },
+    {  80, XK_KP_2 },
+    {  81, XK_Page_Down },
+    {  81, XK_KP_3 },
+    {  81, XK_KP_3 },
+    {  82, XK_Insert },
+    {  82, XK_KP_0 },
+    {  82, XK_KP_0 },
+    {  83, XK_Delete },
+    {  83, XK_KP_Decimal },
+    {  83, XK_KP_Decimal },
+    {  84, 0 },
+    {  84, 0 },
+    {  84, 0 },
+    {  85, 0 },
+    {  85, 0 },
+    {  85, 0 },
+    {  86, 0 },
+    {  86, 0 },
+    {  86, 0 },
+    {  87, XK_F11 },
+    {  87, XK_F23 },
+    {  87, XXX_NOTIMPL(scr11) },
+    {  88, XK_F12 },
+    {  88, XK_F24 },
+    {  88, XXX_NOTIMPL(scr12) },
+    {  89, XK_Return },
+    {  89, XK_Return },
+    {  89, XK_Return },
+    {  90, XK_Control_R },
+    {  90, XK_Control_R },
+    {  90, XK_Control_R },
+    {  91, XK_KP_Divide },
+    {  91, XK_KP_Divide },
+    {  91, XK_KP_Divide },
+    {  92, XXX_NOTIMPL(nscr) },
+    {  92, XXX_NOTIMPL(pscr) },
+    {  92, 0 },
+    {  93, XK_Alt_R },
+    {  93, XK_Alt_R },
+    {  93, XK_Alt_R },
+    {  94, XK_Home },
+    {  94, XK_Home },
+    {  94, XK_Home },
+    {  95, XK_Up },
+    {  95, XK_Up },
+    {  95, XK_Up },
+    {  96, XK_Page_Up },
+    {  96, XK_Page_Up },
+    {  96, XK_Page_Up },
+    {  97, XK_Left },
+    {  97, XK_Left },
+    {  97, XK_Left },
+    {  98, XK_Right },
+    {  98, XK_Right },
+    {  98, XK_Right },
+    {  99, XK_End },
+    {  99, XK_End },
+    {  99, XK_End },
+    { 100, XK_Down },
+    { 100, XK_Down },
+    { 100, XK_Down },
+    { 101, XK_Page_Down },
+    { 101, XK_Page_Down },
+    { 101, XK_Page_Down },
+    { 102, XK_Insert },
+    { 102, XXX_NOTIMPL(paste) },
+    { 102, XK_Insert },
+    { 103, XXX_NOTIMPL(fkey61) },
+    { 103, XXX_NOTIMPL(fkey61) },
+    { 103, XXX_NOTIMPL(fkey61) },
+    { 104, XK_Scroll_Lock },
+    { 104, XXX_NOTIMPL(saver) },
+    { 104, XXX_NOTIMPL(susp) },
+    { 105, XXX_NOTIMPL(fkey62) },
+    { 105, XXX_NOTIMPL(fkey62) },
+    { 105, XXX_NOTIMPL(fkey62) },
+    { 106, XXX_NOTIMPL(fkey63) },
+    { 106, XXX_NOTIMPL(fkey63) },
+    { 106, XXX_NOTIMPL(fkey63) },
+    { 107, XXX_NOTIMPL(fkey64) },
+    { 107, XXX_NOTIMPL(fkey64) },
+    { 107, XXX_NOTIMPL(fkey64) },
+    { 108, 0 },
+    { 108, 0 },
+    { 108, 0 },
+
+    { -1, -1 },
+};
+
+/* current keymap */
+static scancode2keysym_t *scancode2keysym = &us_scancode2keysym[0];
+
+/* AltGr support (0=US101, !0=AltGr) */
+static int flag_need_alternate_graphic = 0;
+
+static int
+parse_keymap_line(char *buf, int *scancode, char *chB, size_t sizeB,
+		  char *chS, size_t sizeS, char *chA, size_t sizeA)
+{
+    const char *p = buf;
+    const char *q;
+    int lockstate = 0;
+
+    /*
+      input format:
+        "#                                                         alt       "
+        "# scan                       cntrl          alt    alt   cntrl lock "
+        "# code  base   shift  cntrl  shift  alt    shift  cntrl  shift state"
+        "  001   esc    esc    esc    esc    esc    esc    debug  esc     O"
+        "  002   '1'    '!'    nop    nop    '1'    '!'    nop    nop     O"
+    */
+
+    /* lockstate */
+    q = strchr(p, '\0');
+    while(q > p && isspace(*--q))
+	;
+#if 0
+    lockstate = *q;
+    if (lockstate == 0)
+	lockstate = 'O';
+#else
+    lockstate = 'O';
+#endif
+
+    /* scan code */
+    while (isspace(*p))
+	p++;
+    if (sscanf(p, "%d", scancode) != 1) {
+	return -1;
+    }
+    while (!isspace(*p))
+	p++;
+
+    /* base */
+    while (isspace(*p))
+	p++;
+    if (*p == '\'' && *(p+1) != '\0' && *(p+2) == '\'') {
+	/* letter */
+	q = p + 3;
+    } else {
+	/* string */
+	q = p;
+	while (!isspace(*q))
+	    q++;
+    }
+    snprintf(chB, sizeB, "%c%.*s", lockstate, q - p, p);
+    p = q;
+
+    /* shift */
+    while (isspace(*p))
+	p++;
+    if (*p == '\'' && *(p+1) != '\0' && *(p+2) == '\'') {
+	/* letter */
+	q = p + 3;
+    } else {
+	/* string */
+	q = p;
+	while (!isspace(*q))
+	    q++;
+    }
+    snprintf(chS, sizeS, "%c%.*s", lockstate, q - p, p);
+    p = q;
+
+    /* cntrl */
+    while (isspace(*p))
+	p++;
+    if (*p == '\'' && *(p+1) != '\0' && *(p+2) == '\'') {
+	/* letter */
+	q = p + 3;
+    } else {
+	/* string */
+	q = p;
+	while (!isspace(*q))
+	    q++;
+    }
+    p = q;
+
+    /* cntrl + shift */
+    while (isspace(*p))
+	p++;
+    if (*p == '\'' && *(p+1) != '\0' && *(p+2) == '\'') {
+	/* letter */
+	q = p + 3;
+    } else {
+	/* string */
+	q = p;
+	while (!isspace(*q))
+	    q++;
+    }
+    p = q;
+
+    /* alt */
+    while (isspace(*p))
+	p++;
+    if (*p == '\'' && *(p+1) != '\0' && *(p+2) == '\'') {
+	/* letter */
+	q = p + 3;
+    } else {
+	/* string */
+	q = p;
+	while (!isspace(*q))
+	    q++;
+    }
+    snprintf(chA, sizeA, "%c%.*s", lockstate, q - p, p);
+    p = q;
+
+    /* alt + shift */
+    while (isspace(*p))
+	p++;
+    if (*p == '\'' && *(p+1) != '\0' && *(p+2) == '\'') {
+	/* letter */
+	q = p + 3;
+    } else {
+	/* string */
+	q = p;
+	while (!isspace(*q))
+	    q++;
+    }
+    p = q;
+
+    /* alt + cntrl */
+    while (isspace(*p))
+	p++;
+    if (*p == '\'' && *(p+1) != '\0' && *(p+2) == '\'') {
+	/* letter */
+	q = p + 3;
+    } else {
+	/* string */
+	q = p;
+	while (!isspace(*q))
+	    q++;
+    }
+    p = q;
+
+    /* alt + cntrl + shift */
+    while (isspace(*p))
+	p++;
+    if (*p == '\'' && *(p+1) != '\0' && *(p+2) == '\'') {
+	/* letter */
+	q = p + 3;
+    } else {
+	/* string */
+	q = p;
+	while (!isspace(*q))
+	    q++;
+    }
+    p = q;
+
+    return 0;
+}
+
+static int
+str2keysym(const char *str)
+{
+    const char *p = str + 1;
+    const char *q = str;
+    char buf[MAX_TMPBUF];
+    char *endptr = NULL;
+    int keysym;
+    int i;
+
+    if (*p == '\'' && *(p+1) != '\0' && *(p+2) == '\'') {
+	/* letter */
+	keysym = (int)*(p+1);
+	if (*q == 'N') {
+	    snprintf(buf, sizeof buf, "NUM_%c", keysym);
+	    p = buf;
+	    goto search_keysym;
+	}
+    } else {
+	/* decnum or hexnum */
+	if (*p == '0' && *(p+1) == 'x') {
+	    keysym = (int)strtoul(p, &endptr, 16);
+	} else {
+	    keysym = (int)strtoul(p, &endptr, 10);
+	}
+	if (*p != '\0' && endptr && *endptr == '\0') {
+	    goto found_keysym;
+	}
+
+	/* string */
+	if (*q == 'N') {
+	    snprintf(buf, sizeof buf, "NUM_%s", p);
+	    p = buf;
+	}
+    search_keysym:
+	keysym = 0;
+	for (i = 0; kbdstr2keysym[i].keysym != -1; i++) {
+	    if (kbdstr2keysym[i].kbdstr == NULL
+		|| kbdstr2keysym[i].kbdstr[0] == '\0')
+		continue;
+	    if (strcasecmp(kbdstr2keysym[i].kbdstr, p) == 0) {
+		keysym = kbdstr2keysym[i].keysym;
+		break;
+	    }
+	}
+    }
+ found_keysym:
+#if 0
+    if (keysym != 0) {
+	printf("found keysym = 0x%x\n", keysym);
+    }
+#endif
+
+    return keysym;
+}
+
+static scancode2keysym_t *
+create_scancode_table(const char *file)
+{
+    scancode2keysym_t *sc2ks = NULL;
+    FILE *fp = NULL;
+    char buf[MAX_TMPBUF];
+    char bufchB[MAX_TMPBUF];
+    char bufchS[MAX_TMPBUF];
+    char bufchA[MAX_TMPBUF];
+    int lineno = 0;
+    int num = 0;
+    int scancode;
+
+    fp = fopen(file, "r");
+    if (fp == NULL) {
+	fprintf(stderr, "open error: %s\n", file);
+	goto error;
+    }
+    flag_need_alternate_graphic = 0;
+    while (fgets(buf, sizeof buf, fp) != NULL) {
+	lineno++;
+	if (buf[0] == '#' || buf[0] =='\n' || buf[0] == '\0')
+	    continue;
+	if (parse_keymap_line(buf, &scancode, bufchB, sizeof bufchB,
+			      bufchS, sizeof bufchS,
+			      bufchA, sizeof bufchA) != 0) {
+	    fprintf(stderr, "parse error at line %d.\n", lineno);
+	    goto error;
+	}
+#if 0
+	printf("scancode=%d, chB=%s, chS=%s, chA=%s\n",
+	       scancode, bufchB, bufchS, bufchA);
+#endif
+	/* allocate for base and shift and alt scancode */
+	num++;
+	if (num == 1) {
+	    sc2ks = malloc(sizeof *sc2ks * (num * 3));
+	    if (sc2ks == NULL) {
+		fprintf(stderr, "malloc error.\n");
+		goto error;
+	    }
+	} else {
+	    sc2ks = realloc(sc2ks, sizeof *sc2ks * (num * 3));
+	    if (sc2ks == NULL) {
+		fprintf(stderr, "realloc error.\n");
+		goto error;
+	    }
+	}
+
+	/* chB */
+	sc2ks[(num - 1) * 3 + 0].scancode = scancode;
+	sc2ks[(num - 1) * 3 + 0].keysym = str2keysym(bufchB);
+	sc2ks[(num - 1) * 3 + 0].modifier = S2KMOD_NONE;
+	/* chS */
+	sc2ks[(num - 1) * 3 + 1].scancode = scancode;
+	sc2ks[(num - 1) * 3 + 1].keysym = str2keysym(bufchS);
+	sc2ks[(num - 1) * 3 + 1].modifier = S2KMOD_SHIFT;
+	/* chA */
+	sc2ks[(num - 1) * 3 + 2].scancode = scancode;
+	sc2ks[(num - 1) * 3 + 2].keysym = str2keysym(bufchA);
+	sc2ks[(num - 1) * 3 + 2].modifier = S2KMOD_ALT;
+
+	/* chB is letter && chB != chA? */
+	if ((sc2ks[(num - 1) * 3 + 0].keysym < 0xff)
+	    && isalpha(sc2ks[(num - 1) * 3 + 0].keysym)
+	    && (sc2ks[(num - 1) * 3 + 0].keysym
+		!= sc2ks[(num - 1) * 3 + 2].keysym)) {
+	    /* AltGr is required */
+	    flag_need_alternate_graphic = 1;
+	}
+    }
+    /* table terminater */
+    if (num == 0) {
+	fprintf(stderr, "no keymap.\n");
+	goto error;
+    } else {
+	sc2ks = realloc(sc2ks, sizeof *sc2ks * ((num * 3) + 1));
+	sc2ks[num * 3].scancode = -1;
+	sc2ks[num * 3].keysym = -1;
+    }
+    fclose(fp);
+    return sc2ks;
+
+error:
+    if (fp != NULL)
+	fclose(fp);
+    if (sc2ks != NULL)
+	free(sc2ks);
+    return NULL;
+}
+
+static void
+destroy_scancode_table(scancode2keysym_t *sc2ks)
+{
+    if (sc2ks == NULL)
+	return;
+    if (sc2ks == &us_scancode2keysym[0])
+	return;
+    free(sc2ks);
+}
+
+int
+load_keymap_table(const char *file)
+{
+    scancode2keysym_t *sc2ks = NULL;
+    char fullpath[FILENAME_MAX];
+
+    if (file == NULL
+	|| strcasecmp(file, "default") == 0) {
+#ifdef DEFAULT_KEYMAP
+	file = DEFAULT_KEYMAP;
+#else
+	sc2ks = &us_scancode2keysym[0];
+	destroy_scancode_table(scancode2keysym);
+	scancode2keysym = sc2ks;
+	return 0;
+#endif /* DEFAULT_KEYMAP */
+    }
+
+    if (strcasecmp(file, "builtin_us") == 0) {
+	sc2ks = &us_scancode2keysym[0];
+    } else {
+	if (file[0] != '/') {
+	    snprintf(fullpath, sizeof fullpath, "%s/%s",
+		     KEYMAPDIR, file);
+	    sc2ks = create_scancode_table(fullpath);
+	} else {
+	    sc2ks = create_scancode_table(file);
+	}
+    }
+    if (sc2ks == NULL) {
+	fprintf(stderr, "create table failed.\n");
+	return -1;
+    }
+    destroy_scancode_table(scancode2keysym);
+    scancode2keysym = sc2ks;
+    return 0;
+}
+
+int
+need_alternate_graphic(void)
+{
+    if (flag_need_alternate_graphic) {
+	/* preferred AltGr */
+	return 1;
+    }
+    /* standard US 101 */
+    return 0;
+}
+
+int
+keysym2scancode(int keysym)
+{
+    return keysym2scancode_modifier(keysym, NULL);
+}
+
+int
+keysym2scancode_modifier(int keysym, int *modifier)
+{
+    int i;
+    
+    for (i = 0; scancode2keysym[i].keysym != -1; i++) {
+	if (scancode2keysym[i].keysym == 0) {
+	    /* nop or notimpl */
+	    continue;
+	}
+	if (scancode2keysym[i].keysym == keysym) {
+	    if (modifier != NULL) {
+		*modifier = scancode2keysym[i].modifier;
+	    }
+	    return scancode2keysym[i].scancode;
+	}
+    }
+    if (modifier != NULL) {
+	*modifier = S2KMOD_NONE;
+    }
+    return -1;
+}
+
+#ifdef TEST
+static void
+dump_keymap_table(const scancode2keysym_t *sc2ks)
+{
+    int i;
+
+    if (sc2ks == NULL)
+	return;
+    printf("keymap table:\n");
+    for (i = 0; sc2ks[i].keysym != -1; i++) {
+	printf("%d: scancode=%d, keysym=%x\n",
+	       i, sc2ks[i].scancode, sc2ks[i].keysym);
+    }
+}
+
+int
+main(int argc, char **argv)
+{
+    char *file = KEYMAPFILE;
+    int scancode;
+
+    if (argc > 1) {
+	file = argv[1];
+    }
+    printf("load: %s\n", file);
+
+    if (load_keymap_table(file) != 0) {
+	fprintf(stderr, "load keymap failed.\n");
+	return 1;
+    }
+#if 0
+    dump_keymap_table(scancode2keysym);
+#endif
+    scancode = keysym2scancode(XK_Delete);
+    printf("XK_Delete = %x\n", scancode);
+    scancode = keysym2scancode('@');
+    printf("'@' = %x\n", scancode);
+    return 0;
+}
+#endif /* TEST */
